﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Samples.ServiceHosting.StorageClient;
using System.Configuration;
using System.Data.Services.Client;
using System.Diagnostics;
using System.Net;

namespace Benday.Azure.Common
{
    public abstract class AzureStorageTable<T> where T : AzureTableStorageEntity, new()
    {
        public AzureStorageTable()
        {
            PartitionKey = CommonConstants.DEFAULT_PARTITION_KEY;
        }

        public AzureStorageTable(string partitionKey)
        {
            if (partitionKey == null)
            {
                throw new ArgumentNullException("partitionKey", "Argument cannot be null.");
            }

            PartitionKey = partitionKey;
        }

        public string PartitionKey { get; set; }

        private static object m_syncRoot = new object();

        private static StorageInformation m_storageInfo;

        public StorageInformation StorageInfo
        {
            get
            {
                if (m_storageInfo == null)
                {
                    lock(m_syncRoot)
                    {
                        if (m_storageInfo == null)
                        {
                            m_storageInfo = new StorageInformation();
                            m_storageInfo.Initialize(GetTableNameForType(typeof(T)));
                        }
                    }
                }

                return m_storageInfo;
            }
        }

        public virtual string GetTableNameForType(Type getNameForThisType)
        {
            return CommonUtility.GetTableNameForType(getNameForThisType);
        }

        private TableStorageDataServiceContext m_serviceContext;
        public TableStorageDataServiceContext ServiceContext
        {
            get
            {
                if (m_serviceContext == null)
                {
                    m_serviceContext = new TableStorageDataServiceContext();
                }
                return m_serviceContext;
            }
        }
        /// <summary>
        /// Return all the folders in the Folder table.
        /// </summary>
        /// <returns></returns>
        public IList<T> GetAll()
        {
            IEnumerable<T> query = from c in ServiceContext.CreateQuery<T>(this.StorageInfo.TableName)
                                   where c.PartitionKey == PartitionKey
                                   select c;

            TableStorageDataServiceQuery<T> tableQuery = new TableStorageDataServiceQuery<T>(
                query as DataServiceQuery<T>, 
                this.StorageInfo.RetrySettings);

            IEnumerable<T> results = tableQuery.ExecuteAllWithRetries();

            return ConvertToList(results);
        }


        protected virtual IList<T> ConvertToList(IEnumerable<T> results)
        {
            if (results == null)
            {
                return null;
            }
            else 
            {
                return new List<T>(results);
            }
        }

        public void Delete(T deleteThis)
        {
            Delete(deleteThis.Id);
        }

        /// <summary>
        /// Remove a row in the table.
        /// </summary>
        /// <param name="id">The Id</param>
        public void Delete(string id)
        {
            try
            {
                IEnumerable<T> query = from c in ServiceContext.CreateQuery<T>(this.StorageInfo.TableName)
                                       where c.PartitionKey == PartitionKey && c.Id == id
                                       select c;

                TableStorageDataServiceQuery<T> q = new TableStorageDataServiceQuery<T>(query as DataServiceQuery<T>, this.StorageInfo.RetrySettings);
                IEnumerable<T> folders = q.ExecuteAllWithRetries();

                int count = 0;
                foreach (T folder in folders)
                {
                    if (count++ > 1)
                    {
                        Log.Write(EventKind.Warning, "There are multiple folders with the same ID {0} in the Folder table. Folder ID should be unique.", id);
                    }

                    ServiceContext.DeleteObject(folder);
                    ServiceContext.SaveChanges();
                }
            }
            catch (InvalidOperationException e)
            {
                Log.Write(EventKind.Warning, "The deletion of the folder (ID={0}) from the data storage failed.", id);
                throw new Exception("Failed to delete the folder ", e);
            }
        }

        protected virtual bool IsUnsaved(T saveThis)
        {
            if (saveThis == null)
            {
                throw new ArgumentNullException("saveThis", "Argument cannot be null.");
            }

            if (saveThis.Id == Guid.Empty.ToString())
            {
                return true;
            }
            else
            {
                return false;
            }
        }

        public void Save(T saveThis)
        {
            if (saveThis == null)
            {
                throw new ArgumentNullException("saveThis", "Argument cannot be null.");
            }

            if (IsUnsaved(saveThis) == true)
            {
                Insert(saveThis);
            }
            else
            {
                Update(saveThis);
            }
        }

        protected virtual void Insert(T saveThis)
        {
            if (saveThis == null)
            {
                throw new ArgumentNullException("saveThis", "Argument cannot be null.");
            }

            saveThis.Id = Guid.NewGuid().ToString();

            try
            {
                ServiceContext.AddObject(StorageInfo.TableName, saveThis);
                ServiceContext.SaveChanges();
            }
            catch (InvalidOperationException e)
            {
                HttpStatusCode httpStatus;
                // when retry policies are used we cannot distinguish between a conflict and success
                // so, in the case of a conflict, we just retrun success here
                if (TableStorageHelpers.EvaluateException(e, out httpStatus) && httpStatus == HttpStatusCode.Conflict)
                {
                    return;
                }

                Log.Write(EventKind.Warning, "Could not add the folder (ID={0}) to the Folder table.", saveThis.Id);
                throw new Exception("Failed to access the Folder table ", e);
            }
        }

        protected virtual void Update(T saveThis)
        {
            if (saveThis == null)
            {
                throw new ArgumentNullException("saveThis", "Argument cannot be null.");
            }

            try
            {
                ServiceContext.UpdateObject(saveThis);
                ServiceContext.SaveChanges();
            }
            catch (InvalidOperationException e)
            {
                Log.Write(EventKind.Warning, "Could not edit the item (ID={0}) in the table.", saveThis.Id);
                throw new Exception("Failed to access the table ", e);
            }
        }

        public T Get(string rowKey)
        {
            return Get(PartitionKey, rowKey, ServiceContext);
        }

        private T Get(string partitionKey, string rowKey, DataServiceContext svc)
        {
            Debug.Assert(svc != null);

            try
            {
                IEnumerable<T> query = from c in svc.CreateQuery<T>(this.StorageInfo.TableName)
                                       where c.PartitionKey == partitionKey &&
                                       c.RowKey == rowKey
                                       select c;

                TableStorageDataServiceQuery<T> q = new TableStorageDataServiceQuery<T>(query as DataServiceQuery<T>, this.StorageInfo.RetrySettings);
                IEnumerable<T> results = q.ExecuteAllWithRetries();

                int count = 0;
                foreach (T resultItem in results)
                {
                    if (count++ > 1)
                    {
                        Log.Write(EventKind.Warning, "There are multiple folders with the same ID (ID={0}) in the table. ID should be unique.", partitionKey);
                    }
                    return resultItem;
                }

                return null;
            }
            catch (InvalidOperationException e)
            {
                Log.Write(EventKind.Warning, "Could not retrieve the folder (ID={0}) from the Folder table.", partitionKey);
                throw new Exception("Failed to access the Folder table ", e);
            }
        }

    }
}
